/*
 * Copyright (C) 2012 Teluu Inc. (http://www.teluu.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <pjsua2/endpoint.hpp>
#include "instant_messaging.hpp"
#include "auth_challenge.hpp"

using namespace pj;

int main(int argc, char *argv[])
{
    try {
        {
            InstantMessagingTests instantMessagingTests;

            instantMessagingTests.immediateResponse();
            instantMessagingTests.deferredResponse();
        }

        {
            AuthChallengeTests authChallengeTests;

            authChallengeTests.immediateResponse();
            authChallengeTests.deferredResponse();
            authChallengeTests.deferredRespondWithCreds();
            authChallengeTests.deferredAbandon();
            authChallengeTests.accountDeleteWithPending();
        }
    } catch (...) {
        return 1;
    }

    return 0;
}

