/*
 * Copyright (C) 2026 Teluu Inc. (http://www.teluu.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef AUTH_CHALLENGE_TESTS_HPP
#define AUTH_CHALLENGE_TESTS_HPP

#include <pjsua2.hpp>
#include <memory>

using namespace pj;

class AuthChallengeTests
{
public:
    AuthChallengeTests();
    AuthChallengeTests(const AuthChallengeTests&) = delete;
    AuthChallengeTests& operator=(const AuthChallengeTests&) = delete;
    ~AuthChallengeTests();

    void immediateResponse();
    void deferredResponse();
    void deferredRespondWithCreds();
    void deferredAbandon();
    void accountDeleteWithPending();

private:
    Endpoint ep;
    pj_uint16_t port;
};

#endif // AUTH_CHALLENGE_TESTS_HPP
