#ifndef PJ_CONFIG_H
#define PJ_CONFIG_H

#if defined(_MSC_VER)
#  pragma warning(disable: 4267) // conversion from 'size_t' to 'int', possible loss of data
#  pragma warning(disable: 4334) // '<<': result of 32-bit shift implicitly converted to 64 bits (was 64-bit shift intended?)
#endif

#endif